from archspec.cpu import host

local_host = host()

__version__ = "0.9.15"

try:
    # pypi wheels won't have these for now
    if local_host.family.name == "aarch64":
        # archspec doesn't detect the relevant features for arm
        # so we assume neon is available,
        # as it's unlikely for a device using the lib to not have it
        from ._etcpak_neon import *  # type: ignore
    elif local_host.family.name == "x86_64":
        if "avx512bw" in local_host.features and "avx512vl" in local_host.features:
            from ._etcpak_avx512 import *  # type: ignore
        elif "avx2" in local_host.features:
            from ._etcpak_avx2 import *  # type: ignore
        elif "sse4_1" in local_host.features:
            from ._etcpak_sse41 import *  # type: ignore
except ImportError:
    pass

if "compress_bc1" not in locals():
    from ._etcpak_none import *  # type: ignore

# legacy mappings for backwards compatibility
compress_to_dxt1 = compress_bc1  # noqa: F405
compress_to_dxt1_dither = compress_bc1_dither  # noqa: F405
compress_to_dxt5 = compress_bc3  # noqa: F405
compress_to_etc1 = compress_etc1_rgb  # noqa: F405
compress_to_etc1_dither = compress_etc1_rgb_dither  # noqa: F405
compress_to_etc2 = compress_etc2_rgb  # noqa: F405
compress_to_etc2_rgba = compress_etc2_rgba  # noqa: F405
decode_dxt1 = decompress_bc1  # noqa: F405
decode_dxt5 = decompress_bc3  # noqa: F405
decode_etc_rgba = decompress_etc2_rgba  # noqa: F405


def compress_to_etc1_alpha(data: bytes, width: int, height: int) -> bytes:
    raise NotImplementedError("This function was removed in etcpak 0.9.9")


def compress_to_etc2_alpha(data: bytes, width: int, height: int) -> bytes:
    raise NotImplementedError("This function was removed in etcpak 0.9.9")


def set_use_heuristics(use_heuristics: bool) -> None:
    raise NotImplementedError("This function was removed in etcpak 0.9.9")


def get_use_multithreading() -> bool:
    raise NotImplementedError("This function was removed in etcpak 0.9.9")


__all__ = (
    "__version__",
    "compress_bc1",
    "compress_bc1_dither",
    "compress_bc3",
    "compress_bc4",
    "compress_bc5",
    "compress_bc7",
    "compress_etc1_rgb",
    "compress_etc1_rgb_dither",
    "compress_etc2_rgb",
    "compress_etc2_rgba",
    "compress_eac_r",
    "compress_eac_rg",
    "decompress_etc1_rgb",
    "decompress_etc2_rgb",
    "decompress_etc2_rgba",
    "decompress_etc2_r11",
    "decompress_etc2_rg11",
    "decompress_bc1",
    "decompress_bc3",
    "decompress_bc4",
    "decompress_bc5",
    "decompress_bc7",
    "BC7CompressBlockParams",
    # legacy mappings for backwards compatibility
    "compress_to_dxt1",
    "compress_to_dxt1_dither",
    "compress_to_dxt5",
    "compress_to_etc1",
    "compress_to_etc1_dither",
    "compress_to_etc2",
    "compress_to_etc2_rgba",
    "decode_dxt1",
    "decode_dxt5",
    "decode_etc_rgba",
    "compress_to_etc1_alpha",
    "compress_to_etc2_alpha",
    "set_use_heuristics",
    "get_use_multithreading",
)



import os
import etcpak
import texture2ddecoder
from PIL import Image
import imagehash


def _test_helper(
    name: str, channels: int, encode_func: callable, decode_func: callable
):
    p = os.path.dirname(os.path.realpath(__file__))
    for fp in os.listdir(p):
        if not fp[-4:] in [".png", ".jpg", ".bmp", "jpeg"]:
            continue
        img = Image.open(os.path.join(p, fp)).convert("RGBA")
        img = img.crop((0, 0, img.width - img.width % 4, img.height - img.height % 4))

        raw = img.tobytes()
        # compress
        data = encode_func(raw, img.width, img.height)
        # decompress
        dimg = Image.frombytes(
            "RGBA", img.size, decode_func(data, img.width, img.height), "raw", "BGRA"
        )

        # compare to original
        if channels in [1, 2]:
            # todo
            continue
        elif channels == 3:
            img = img.convert("RGB")
            dimg = dimg.convert("RGB")

        assert compare_images(img, dimg), f"Failed {name}: {fp}"


def compare_images(im1: Image.Image, im2: Image.Image) -> bool:
    # lossy compression, so some leeway is allowed
    return abs(imagehash.average_hash(im1) - imagehash.average_hash(im2)) <= 1


def test_bc1():
    _test_helper("BC1", 3, etcpak.compress_bc1, texture2ddecoder.decode_bc1)


def test_bc1_dither():
    _test_helper(
        "BC1 Dither", 3, etcpak.compress_bc1_dither, texture2ddecoder.decode_bc1
    )


def test_bc3():
    _test_helper("BC3", 4, etcpak.compress_bc3, texture2ddecoder.decode_bc3)


def test_bc4():
    _test_helper("BC4", 1, etcpak.compress_bc4, texture2ddecoder.decode_bc4)


def test_bc5():
    _test_helper("BC5", 3, etcpak.compress_bc5, texture2ddecoder.decode_bc5)


def test_bc7():
    _test_helper("BC7", 4, etcpak.compress_bc7, texture2ddecoder.decode_bc7)


def test_etc1_rgb():
    _test_helper("ETC1 RGB", 3, etcpak.compress_etc1_rgb, texture2ddecoder.decode_etc1)


def test_etc1_rgb_dither():
    _test_helper(
        "ETC1 RGB Dither",
        3,
        etcpak.compress_etc1_rgb_dither,
        texture2ddecoder.decode_etc1,
    )


def test_etc2_rgb():
    _test_helper("ETC2 RGB", 3, etcpak.compress_etc2_rgb, texture2ddecoder.decode_etc2)


def test_etc2_rgba():
    _test_helper(
        "ETC2 RGBA", 4, etcpak.compress_etc2_rgba, texture2ddecoder.decode_etc2a8
    )


if __name__ == "__main__":
    for name in dir():
        if name.startswith("test_"):
            globals()[name]()



from __future__ import annotations

import os
from typing import List

from archspec.cpu import host
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext
from wheel.bdist_wheel import bdist_wheel

ETCPAK_SOURCES = [
    "bc7enc.cpp",
    # "bcdec.c", mac issue, fixed by moving define to compiler define arg
    "Decode.cpp",
    "ColorSpace.cpp",
    "Debug.cpp",
    "Dither.cpp",
    "Error.cpp",
    "mmap.cpp",
    "ProcessDxtc.cpp",
    "ProcessRGB.cpp",
    "System.cpp",
    "Tables.cpp",
    "TaskDispatch.cpp",
    "Timing.cpp",
]

ETCPAK_HEADERS = [
    "b7enc.h",
    "bcdec.h",
    "Decode.hpp",
    "ColorSpace.hpp",
    "DataProvider.hpp",
    "Debug.hpp",
    "Dither.hpp",
    "Error.hpp",
    "ForceInline.hpp",
    "Math.hpp",
    "MipMap.hpp",
    "mmap.hpp",
    "ProcessCommon.hpp",
    "ProcessDxtc.hpp",
    "ProcessRGB.hpp",
    "Semaphore.hpp",
    "System.hpp",
    "Tables.hpp",
    "TaskDispatch.hpp",
    "Timing.hpp",
    "Vector.hpp",
]

class BuildConfig:
    __SSE4_1__: bool = False
    __AVX2__: bool = False
    __AVX512BW__: bool = False
    __AVX512VL__: bool = False
    __AVX512F__: bool = False
    __ARM_NEON: bool = False
    msvc_flags: List[str] = []
    unix_flags: List[str] = []

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
    
    def items(self):
        return self.__dict__.items()

configs = {
    "sse41": BuildConfig(
        __SSE4_1__=True,
        msvc_flags=["/arch:SSE4.1"],
        unix_flags=["-msse4.1"]
    ),
    "avx2": BuildConfig(
        __SSE4_1__=True,
        __AVX2__=True,
        msvc_flags=["/arch:AVX2"],
        unix_flags=["-msse4.1", "-mfma", "-mbmi", "-mavx2"]
    ),
    "avx512": BuildConfig(
        __SSE4_1__=True,
        __AVX2__=True,
        __AVX512BW__=True,
        __AVX512VL__=True,
        msvc_flags=["/arch:AVX512"],
        unix_flags=["-msse4.1", "-mfma", "-mbmi", "-mavx2", "-mavx512vl", "-mavx512bw", "-mavx512dq", "-mavx512f"]
    ),
    "neon": BuildConfig(
        __ARM_NEON=True,
    )
}


class CustomBuildExt(build_ext):
    def build_extensions(self):
        extra_compile_args: List[str] = []
        extra_link_args: List[str] = []

        # check for cibuildwheel
        cibuildwheel = os.environ.get("CIBUILDWHEEL", False)

        msvc: bool = False
        if self.compiler.compiler_type == "msvc":
            extra_compile_args = [
                "/std:c++20",
                "/Zc:strictStrings-",
                "/DNOMINMAX",
                "/GL",
            ]
            extra_link_args = ["/LTCG:incremental"]
            msvc = True
        else:
            extra_compile_args = [
                "-std=c++20",
                "-flto",
            ]
            extra_link_args = ["-flto"]

            if not cibuildwheel:
                # do some native optimizations for the current machine
                # can't be used for generic builds
                if "-arm" in self.plat_name or "-aarch64" in self.plat_name:
                    native_arg = "-mcpu=native"
                else:
                    native_arg = "-march=native"
                extra_compile_args.append(native_arg)

        local_host = host()

        if self.plat_name.endswith(("amd64", "x86_64")):
            if cibuildwheel:
                self.extensions.extend(
                    [
                        ETCPAKExtension("none", BuildConfig()),
                        ETCPAKExtension("sse41", configs["sse41"]),
                        ETCPAKExtension("avx2", configs["avx2"]),
                        ETCPAKExtension("avx512", configs["avx512"]),
                    ]
                )
            elif "avx512bw" in local_host.features and "avx512vl" in local_host.features:
                self.extensions.append(ETCPAKExtension("avx512", configs["avx512"]))
            elif "avx2" in local_host.features:
                self.extensions.append(ETCPAKExtension("avx2", configs["avx2"]))
            elif "sse4_1" in local_host.features:
                self.extensions.append(ETCPAKExtension("sse41", configs["sse41"]))

        elif self.plat_name.endswith(("arm64", "aarch64")):
            # TODO: somehow detect neon, sve128, sve256
            # atm assume neon is always available
            self.extensions.append(ETCPAKExtension("neon", configs["neon"]))
        elif self.plat_name.endswith("armv7l"):
            # TODO: detect neon
            pass

        for ext in self.extensions:
            ext: ETCPAKExtension
            ext.extra_compile_args.extend(extra_compile_args)
            ext.extra_link_args.extend(extra_link_args)

            build_config = ext.build_config
            if msvc:
                ext.extra_compile_args.extend(build_config.msvc_flags)
            else:
                ext.extra_compile_args.extend(build_config.unix_flags)
            
            for key, value in build_config.items():
                if key.startswith("__"):
                    if value:
                        ext.define_macros.append((key, None))
                    else:
                        ext.undef_macros.append(key)

        super().build_extensions()



class ETCPAKExtension(Extension):
    build_config: BuildConfig
    _needs_stub: bool = False

    def __init__(self, name: str, build_config: BuildConfig):
        module_name = f"_etcpak_{name}"
        super().__init__(
            f"etcpak.{module_name}",
            sources=[
            "src/pylink.cpp",
            *[f"src/etcpak/{src}" for src in ETCPAK_SOURCES],
            ],
            depends=[
                "src/pybc7params.hpp",
                *[
                    f"src/etcpak/{header}"
                    for header in ETCPAK_HEADERS
                ],
            ],
            include_dirs=["src/etcpak"],
            language="c++",
            extra_compile_args=[
                "-DNDEBUG",
                "-DNO_GZIP",
            ],
            define_macros=[
                ("Py_LIMITED_API", "0x03070000"),
                ("MODULE_NAME", '"module_name"'),
                ("INIT_FUNC_NAME", f"PyInit_{module_name}"),
                # Mac fix due to .c problem
                ("BCDEC_IMPLEMENTATION", "1"),
            ],
            py_limited_api=True,
        )

        self.build_config = build_config



class bdist_wheel_abi3(bdist_wheel):
    def get_tag(self):
        python, abi, plat = super().get_tag()

        if python.startswith("cp"):
            # on CPython, our wheels are abi3 and compatible back to 3.6
            return "cp37", "abi3", plat

        return python, abi, plat

setup(
    name="etcpak",
    packages=["etcpak"],
    package_data={"etcpak": ["*.py", "*.pyi", "py.typed"]},
    ext_modules=[ETCPAKExtension("none", BuildConfig())],
    cmdclass={"build_ext": CustomBuildExt, "bdist_wheel": bdist_wheel_abi3},
)


from typing import Optional

class BC7CompressBlockParams:
    """BC7 compression block parameters.

    Attributes
    ----------
        m_mode_mask: int
            The mode mask.
        m_max_partitions: int
            The maximum partitions.
            m_max_partitions may range from 0 (disables mode 1) to BC7ENC_MAX_PARTITIONS (64).
            The higher this value, the slower the compressor, but the higher the quality.
        m_weights: List[int]
            The weights.
            Relative RGBA or YCbCrA weights.
        m_uber_level: int
            The uber level.
            m_uber_level may range from 0 to BC7ENC_MAX_UBER_LEVEL (4).
            The higher this value, the slower the compressor, but the higher the quality.
        m_perceptual: bool
            Perceptual.
            If m_perceptual is true, colorspace error is computed in YCbCr space, otherwise RGB.
        m_try_least_squares: bool
            Try least squares.
            Set m_try_least_squares to false for slightly faster/lower quality compression.
        m_mode17_partition_estimation_filterbank: bool
            Mode 17 partition estimation filterbank.
            When m_mode17_partition_estimation_filterbank,
            the mode1 partition estimator skips lesser used partition patterns unless they are strongly predicted to be potentially useful.
            There's a slight loss in quality with this enabled (around .08 dB RGB PSNR or .05 dB Y PSNR),
            but up to a 11% gain in speed depending on the other settings.
        m_force_selectors: bool
            Force selectors.
        m_force_alpha: bool
            Force alpha.
        m_quant_mode6_endpoints: bool
            Quant mode 6 endpoints.
        m_bias_mode1_pbits: bool
            Bias mode 1 pbits.
        m_pbit1_weight: float
            Pbit 1 weight.
        m_mode1_error_weight: float
            Mode 1 error weight.
        m_mode5_error_weight: float
            Mode 5 error weight.
        m_mode6_error_weight: float
            Mode 6 error weight.
        m_mode7_error_weight: float
            Mode 7 error weight.
    """

    m_mode_mask: int = 2**32 - 1
    m_max_partitions: int = 64
    m_try_least_squares: bool = True
    m_mode17_partition_estimation_filterbank: bool = True
    m_uber_level: int = 0
    m_force_selectors: bool = False
    m_force_alpha: bool = False
    m_quant_mode6_endpoints: bool = False
    m_bias_mode1_pbits: bool = False
    m_pbit1_weigh: float = 1.0
    m_mode1_error_weight: float = 1.0
    m_mode5_error_weight: float = 1.0
    m_mode6_error_weight: float = 1.0
    m_mode7_error_weight: float = 1.0
    m_perceptual: bool = False
    m_weights: list[int] = [1, 1, 1, 1]

    def __init__(self) -> None: ...
    def init_linear_weights(self) -> None:
        """Initialize the weights for linear mode.

        m_perceptual: false
        m_weights = [1,1,1,1]
        """
        ...
    def init_perceptual_weights(self) -> None:
        """Initialize the weights for perceptual mode.

        m_perceptual: true
        m_weights = [128,64,16,32]
        """

def compress_bc1(data: bytes, width: int, height: int) -> bytes: ...
def compress_bc1_dither(data: bytes, width: int, height: int) -> bytes: ...
def compress_bc3(data: bytes, width: int, height: int) -> bytes: ...
def compress_bc4(data: bytes, width: int, height: int) -> bytes: ...
def compress_bc5(data: bytes, width: int, height: int) -> bytes: ...
def compress_bc7(
    data: bytes, width: int, height: int, params: Optional[BC7CompressBlockParams]
) -> bytes: ...
def compress_etc1_rgb(data: bytes, width: int, height: int) -> bytes: ...
def compress_etc1_rgb_dither(data: bytes, width: int, height: int) -> bytes: ...
def compress_etc2_rgb(data: bytes, width: int, height: int) -> bytes: ...
def compress_etc2_rgba(data: bytes, width: int, height: int) -> bytes: ...
def compress_eac_r(data: bytes, width: int, height: int) -> bytes: ...
def compress_eac_rg(data: bytes, width: int, height: int) -> bytes: ...
def decompress_etc1_rgb(data: bytes, width: int, height: int) -> bytes: ...
def decompress_etc2_rgb(data: bytes, width: int, height: int) -> bytes: ...
def decompress_etc2_rgba(data: bytes, width: int, height: int) -> bytes: ...
def decompress_etc2_r11(data: bytes, width: int, height: int) -> bytes: ...
def decompress_etc2_rg11(data: bytes, width: int, height: int) -> bytes: ...
def decompress_bc1(data: bytes, width: int, height: int) -> bytes: ...
def decompress_bc3(data: bytes, width: int, height: int) -> bytes: ...
def decompress_bc4(data: bytes, width: int, height: int) -> bytes: ...
def decompress_bc5(data: bytes, width: int, height: int) -> bytes: ...
def decompress_bc7(data: bytes, width: int, height: int) -> bytes: ...
